@@ -127,6 +127,12 @@ typedef struct
    */
   guint failed : 1;
 
+  /*
+   * Only set once we've emitted the ::failed signal (so we only do that
+   * a single time).
+   */
+  guint emitted_failed : 1;
+
   /*
    * If we should try to use gvariant encoding when communicating with
    * our peer. This is helpful to be able to lower parser and memory
@@ -266,6 +272,22 @@ cancel_pending_from_main (JsonrpcClient *self,
   priv->invocations = g_hash_table_new_full (NULL, NULL, NULL, g_object_unref);
 }
 
+static gboolean
+emit_failed_from_main (JsonrpcClient *self)
+{
+  JsonrpcClientPrivate *priv = jsonrpc_client_get_instance_private (self);
+
+  g_assert (JSONRPC_IS_CLIENT (self));
+
+  if (!priv->emitted_failed)
+    {
+      priv->emitted_failed = TRUE;
+      g_signal_emit (self, signals [FAILED], 0);
+    }
+
+  return G_SOURCE_REMOVE;
+}
+
 /*
  * jsonrpc_client_panic:
  *
@@ -297,6 +319,16 @@ jsonrpc_client_panic (JsonrpcClient *self,
    */
   g_clear_object (&priv->input_stream);
   g_clear_object (&priv->output_stream);
+
+  /*
+   * Queue a "failed" signal from a main loop callback so that we don't
+   * get the client into weird stuff from signal callbacks here.
+   */
+  if (!priv->emitted_failed)
+    g_idle_add_full (G_MAXINT,
+                     (GSourceFunc)emit_failed_from_main,
+                     g_object_ref (self),
+                     g_object_unref);
 }
 
 /*
@@ -847,7 +879,7 @@ jsonrpc_client_call_sync_cb (GObject      *object,
  * @method: the name of the method to call
  * @params: (transfer none) (nullable): A #GVariant of parameters or %NULL
  * @cancellable: (nullable): A #GCancellable or %NULL
- * @return_value: (nullable) (out): A location for a #JsonNode.
+ * @return_value: (nullable) (out): A location for a #GVariant
  *
  * Synchronously calls @method with @params on the remote peer.
  *
@@ -902,11 +934,106 @@ jsonrpc_client_call (JsonrpcClient  *self,
   return ret;
 }
 
+/**
+ * jsonrpc_client_call_with_id_async:
+ * @self: A #JsonrpcClient
+ * @method: the name of the method to call
+ * @params: (transfer none) (nullable): A #GVariant of parameters or %NULL
+ * @id: (out) (transfer full) (optional): a location for a #GVariant
+ *   describing the identifier used for the method call, or %NULL.
+ * @cancellable: (nullable): A #GCancellable or %NULL
+ * @callback: a callback to executed upon completion
+ * @user_data: user data for @callback
+ *
+ * Asynchronously calls @method with @params on the remote peer.
+ *
+ * Upon completion or failure, @callback is executed and it should
+ * call jsonrpc_client_call_finish() to complete the request and release
+ * any memory held.
+ *
+ * This function is similar to jsonrpc_client_call_async() except that
+ * it allows the caller to get the id of the command which might be useful
+ * in systems where you can cancel the operation (such as the Language
+ * Server Protocol).
+ *
+ * If @params is floating, the floating reference is consumed.
+ *
+ * Since: 3.30
+ */
+void
+jsonrpc_client_call_with_id_async (JsonrpcClient       *self,
+                                   const gchar         *method,
+                                   GVariant            *params,
+                                   GVariant           **id,
+                                   GCancellable        *cancellable,
+                                   GAsyncReadyCallback  callback,
+                                   gpointer             user_data)
+{
+  JsonrpcClientPrivate *priv = jsonrpc_client_get_instance_private (self);
+  g_autoptr(GVariant) message = NULL;
+  g_autoptr(GTask) task = NULL;
+  g_autoptr(GError) error = NULL;
+  GVariantDict dict;
+  gint64 idval;
+
+  g_return_if_fail (JSONRPC_IS_CLIENT (self));
+  g_return_if_fail (method != NULL);
+  g_return_if_fail (!cancellable || G_IS_CANCELLABLE (cancellable));
+
+  if (id != NULL)
+    *id = NULL;
+
+  task = g_task_new (self, cancellable, callback, user_data);
+  g_task_set_source_tag (task, jsonrpc_client_call_async);
+
+  if (!jsonrpc_client_check_ready (self, &error))
+    {
+      g_task_return_error (task, g_steal_pointer (&error));
+      return;
+    }
+
+  g_signal_connect_object (task,
+                           "notify::completed",
+                           G_CALLBACK (jsonrpc_client_call_notify_completed),
+                           self,
+                           G_CONNECT_SWAPPED);
+
+  idval = ++priv->sequence;
+
+  g_task_set_task_data (task, GINT_TO_POINTER (idval), NULL);
+
+  /* Use empty maybe type for NULL, and floating reference will
+   * be consumed by g_variant_dict_insert_value() below. */
+  if (params == NULL)
+    params = g_variant_new_maybe (G_VARIANT_TYPE_VARIANT, NULL);
+
+  g_variant_dict_init (&dict, NULL);
+  g_variant_dict_insert (&dict, "jsonrpc", "s", "2.0");
+  g_variant_dict_insert (&dict, "id", "x", idval);
+  g_variant_dict_insert (&dict, "method", "s", method);
+  g_variant_dict_insert_value (&dict, "params", params);
+  message = g_variant_dict_end (&dict);
+
+  g_hash_table_insert (priv->invocations, GINT_TO_POINTER (idval), g_object_ref (task));
+
+  jsonrpc_output_stream_write_message_async (priv->output_stream,
+                                             message,
+                                             cancellable,
+                                             jsonrpc_client_call_write_cb,
+                                             g_steal_pointer (&task));
+
+  if (priv->is_first_call)
+    jsonrpc_client_start_listening (self);
+
+  if (id != NULL)
+    *id = g_variant_take_ref (g_variant_new_int64 (idval));
+}
+
 /**
  * jsonrpc_client_call_async:
  * @self: A #JsonrpcClient
  * @method: the name of the method to call
- * @params: (transfer none) (nullable): A #JsonNode of parameters or %NULL
+ * @params: (transfer none) (nullable): A #GVariant of parameters or %NULL
  * @cancellable: (nullable): A #GCancellable or %NULL
  * @callback: a callback to executed upon completion
  * @user_data: user data for @callback
@@ -929,65 +1056,14 @@ jsonrpc_client_call_async (JsonrpcClient       *self,
                            GAsyncReadyCallback  callback,
                            gpointer             user_data)
 {
-  JsonrpcClientPrivate *priv = jsonrpc_client_get_instance_private (self);
-  g_autoptr(GVariant) message = NULL;
-  g_autoptr(GTask) task = NULL;
-  g_autoptr(GError) error = NULL;
-  GVariantDict dict;
-  gint64 id;
-
-  g_return_if_fail (JSONRPC_IS_CLIENT (self));
-  g_return_if_fail (method != NULL);
-  g_return_if_fail (!cancellable || G_IS_CANCELLABLE (cancellable));
-
-  task = g_task_new (self, cancellable, callback, user_data);
-  g_task_set_source_tag (task, jsonrpc_client_call_async);
-
-  if (!jsonrpc_client_check_ready (self, &error))
-    {
-      g_task_return_error (task, g_steal_pointer (&error));
-      return;
-    }
-
-  g_signal_connect_object (task,
-                           "notify::completed",
-                           G_CALLBACK (jsonrpc_client_call_notify_completed),
-                           self,
-                           G_CONNECT_SWAPPED);
-
-  id = ++priv->sequence;
-
-  g_task_set_task_data (task, GINT_TO_POINTER (id), NULL);
-
-  /* Use empty maybe type for NULL, and floating reference will
-   * be consumed by g_variant_dict_insert_value() below. */
-  if (params == NULL)
-    params = g_variant_new_maybe (G_VARIANT_TYPE_VARIANT, NULL);
-
-  g_variant_dict_init (&dict, NULL);
-  g_variant_dict_insert (&dict, "jsonrpc", "s", "2.0");
-  g_variant_dict_insert (&dict, "id", "x", id);
-  g_variant_dict_insert (&dict, "method", "s", method);
-  g_variant_dict_insert_value (&dict, "params", params);
-  message = g_variant_dict_end (&dict);
-
-  g_hash_table_insert (priv->invocations, GINT_TO_POINTER (id), g_object_ref (task));
-
-  jsonrpc_output_stream_write_message_async (priv->output_stream,
-                                             message,
-                                             cancellable,
-                                             jsonrpc_client_call_write_cb,
-                                             g_steal_pointer (&task));
-
-  if (priv->is_first_call)
-    jsonrpc_client_start_listening (self);
+  jsonrpc_client_call_with_id_async (self, method, params, NULL, cancellable, callback, user_data);
 }
 
 /**
  * jsonrpc_client_call_finish:
  * @self: A #JsonrpcClient.
  * @result: A #GAsyncResult provided to the callback in jsonrpc_client_call_async()
- * @return_value: (out) (nullable): A location for a #JsonNode or %NULL
+ * @return_value: (out) (nullable): A location for a #GVariant or %NULL
  * @error: a location for a #GError or %NULL
  *
  * Completes an asynchronous call to jsonrpc_client_call_async().
@@ -1002,7 +1078,7 @@ jsonrpc_client_call_finish (JsonrpcClient  *self,
                             GVariant      **return_value,
                             GError        **error)
 {
-  g_autoptr(JsonNode) local_return_value = NULL;
+  g_autoptr(GVariant) local_return_value = NULL;
   gboolean ret;
 
   g_return_val_if_fail (JSONRPC_IS_CLIENT (self), FALSE);
@@ -1230,7 +1306,7 @@ jsonrpc_client_close (JsonrpcClient  *self,
                                      "The underlying stream was closed");
   cancel_pending_from_main (self, local_error);
 
-  g_signal_emit (self, signals [FAILED], 0);
+  emit_failed_from_main (self);
 
   return ret;
 }
@@ -1395,7 +1471,7 @@ jsonrpc_client_reply_error_finish (JsonrpcClient  *self,
  * jsonrpc_client_reply:
  * @self: a #JsonrpcClient
  * @id: (transfer none) (not nullable): the id of the message to reply
- * result: (transfer none) (nullable): the return value or %NULL
+ * @result: (transfer none) (nullable): the return value or %NULL
  * @cancellable: (nullable): a #GCancellable, or %NULL
  * @error: a #GError, or %NULL
  *
